@@ -42,20 +42,33 @@
  *
  * AFF and AFFLIB is a trademark of Simson Garfinkel and Basis Technology Corp.
  */
-#include <ctype.h>
+#include "affconfig.h"
 #include "afflib.h"
 #include "afflib_i.h"
 #include "vnode_split_raw.h"
 
+#ifdef HAVE_CTYPE_H
+#include <ctype.h>
+#endif
+
+#ifndef HAVE_ISDIGIT
+static int isdigit(char ch)
+{
+    return ch>='0' && ch<='9';
+}
+#endif
+
+
+
 /* split raw file implementation with optional metadata support */
 struct split_raw_private {
-    uint64 num_raw_files;			// number of raw files
-    int    *fds;				// array of file descriptors for each open raw file
-    uint64 *pos;				// where we are in each file
+    uint   num_raw_files;	// number of raw files
+    int    *fds;		// array of file descriptors for each open raw file
+    uint64_t *pos;		// where we are in each file
     char *first_raw_fname;    /* The filename of the first raw file. */
     char *next_raw_fname;     /* The filename of the next raw file, or 0
 				 when one big file is used. */
-    int64 cur_page; // current page number, used for split_raw_get_next_seg
+    int64_t cur_page; // current page number, used for split_raw_get_next_seg
 };
 
 static inline struct split_raw_private *SPLIT_RAW_PRIVATE(AFFILE *af)
@@ -74,13 +87,14 @@ static inline const char *last4(const char *filename)
 }
 
 /* Return 1 if a file is a split raw file... */
-static int split_raw_identify_file(const char *filename)
+static int split_raw_identify_file(const char *filename,int exists)
 {
-  const char *l4 = last4 (filename);
+    if(exists && access(filename,R_OK)!=0) return 0;	// needs to exist and it doesn't
+    const char *l4 = last4 (filename);
 
-  return (!strcmp (l4, ".000"))
-      || (!strcmp (l4, ".001"))
-      || (!strcasecmp (l4, ".aaa"));
+    return (!strcmp (l4, ".000"))
+	|| (!strcmp (l4, ".001"))
+	|| (!strcasecmp (l4, ".aaa"));
 }
 
 /* split_raw_close:
@@ -91,13 +105,13 @@ static int split_raw_close(AFFILE *af)
 {
     struct split_raw_private *srp = SPLIT_RAW_PRIVATE(af);
 
-    for (uint64 i = 0; i < srp->num_raw_files; i++){
+    for (uint64_t i = 0; i < srp->num_raw_files; i++){
 	close (srp->fds[i]);
     }
     if (srp->fds)    free (srp->fds);
     if (srp->pos)    free (srp->pos);
     if (srp->first_raw_fname) free (srp->first_raw_fname);
-    if (srp->next_raw_fname) free (srp->next_raw_fname);
+    if (srp->next_raw_fname)  free (srp->next_raw_fname);
     free(srp);
     af->vnodeprivate = 0;
     return 0;
@@ -169,18 +183,18 @@ static void srp_add_fd(AFFILE *af,int fd)
     srp->num_raw_files++;
     srp->fds = (int *)realloc (srp->fds, sizeof (int) * (srp->num_raw_files));
     srp->fds[srp->num_raw_files - 1] = fd;
-    srp->pos = (uint64 *)realloc (srp->pos, sizeof (uint64) * (srp->num_raw_files));
+    srp->pos = (uint64_t *)realloc (srp->pos, sizeof (uint64_t) * (srp->num_raw_files));
     srp->pos[srp->num_raw_files - 1] = 0;
 }
 
 
-static int split_raw_open_internal(AFFILE *af, uint64 *image_size)
+static int split_raw_open_internal(AFFILE *af, uint64_t *image_size)
 {
     struct split_raw_private *srp = SPLIT_RAW_PRIVATE(af);
     int fd;
     struct stat sb;
 
-    fd = open(srp->first_raw_fname, af->openflags, 0666);
+    fd = open(srp->first_raw_fname, af->openflags|O_BINARY, 0666);
     if (fd < 0) {
       (*af->error_reporter)("split_raw_open_internal: open(%s): ",af->fname);
 	return -1;
@@ -189,7 +203,7 @@ static int split_raw_open_internal(AFFILE *af, uint64 *image_size)
     srp->num_raw_files = 1;
     srp->fds = (int *)malloc (sizeof (int));
     srp->fds[0] = fd;
-    srp->pos = (uint64 *)malloc (sizeof (uint64));
+    srp->pos = (uint64_t *)malloc (sizeof (uint64_t));
     if (fstat (fd, &sb) != 0) {
       (*af->error_reporter)("split_raw_open_internal: fstat(%s): ",af->fname);
 	close (fd);
@@ -218,7 +232,7 @@ static int split_raw_open_internal(AFFILE *af, uint64 *image_size)
 	    return -1;
 	}
 	fd = open(srp->next_raw_fname,
-		   af->openflags & O_RDWR ? O_RDWR : O_RDONLY);
+		   af->openflags & O_RDWR ? (O_RDWR|O_BINARY) : (O_RDONLY|O_BINARY));
 
 	if (fd < 0) {
 	    if (errno != ENOENT) {
@@ -246,7 +260,7 @@ static int split_raw_open_internal(AFFILE *af, uint64 *image_size)
 	  (*af->error_reporter)("split_raw_open_internal: fstat(%s): ",af->fname);
 	    return -1;
 	}
-	if ((uint64)sb.st_size != af->maxsize){
+	if ((uint64_t)sb.st_size != af->maxsize){
 	    current_file_must_be_last = 1;
 	}
     } while (1);
@@ -293,19 +307,20 @@ static int split_raw_vstat(AFFILE *af,struct af_vnode_info *vni)
     vni->pagesize  = af->image_pagesize;
     vni->supports_compression = 0;
     vni->supports_metadata    = 0;
+    vni->is_raw               = 1;
     vni->changable_pagesize   = 1;	// change it at any time
     vni->changable_sectorsize = 1;	// change it at any time
     return 0;
 }
 
-static int split_raw_read(AFFILE *af, unsigned char *buf, uint64 pos,int count)
+static int split_raw_read(AFFILE *af, unsigned char *buf, uint64_t pos,size_t count)
 {
     struct split_raw_private *srp = SPLIT_RAW_PRIVATE(af);
     off_t c3;
     int ret = 0;				// how many bytes read
 
     if ((af->image_size - pos) < (unsigned)count){
-	count = af->image_size - pos;
+	count = (off_t)(af->image_size - pos);
     }
 
     while (count > 0) {
@@ -313,11 +328,11 @@ static int split_raw_read(AFFILE *af, unsigned char *buf, uint64 pos,int count)
 	off_t file_offset = 0;
 	
 	if (af->maxsize) {		// if we do file segments
-	    filenum  = pos / af->maxsize;
-	    file_offset = pos % af->maxsize;
+	    filenum     = (int)(pos / af->maxsize);
+	    file_offset = (off_t)(pos % af->maxsize);
 	} else {
-	    filenum  = 0;
-	    file_offset = pos;
+	    filenum     = 0;
+	    file_offset = (off_t)pos;
 	}
 	if (file_offset != (off_t) srp->pos[filenum]) {
 	    off_t c2 = lseek (srp->fds[filenum], file_offset, SEEK_SET);
@@ -328,7 +343,7 @@ static int split_raw_read(AFFILE *af, unsigned char *buf, uint64 pos,int count)
 	    srp->pos[filenum] = c2;	// this file starts here
 	}
 	if (af->maxsize && ((af->maxsize - file_offset) < (unsigned) count))
-	    c3 = af->maxsize - file_offset;
+	    c3 = (off_t)(af->maxsize - file_offset);
 	else
 	    c3 = count;
 	off_t c4 = read (srp->fds[filenum], buf, c3);
@@ -352,7 +367,7 @@ static int split_raw_read(AFFILE *af, unsigned char *buf, uint64 pos,int count)
  * and just seek to the last character and write a single NUL.
  */
 
-int split_raw_write_internal2(AFFILE *af, unsigned char *buf, uint64 pos,int count)
+int split_raw_write_internal2(AFFILE *af, unsigned char *buf, uint64_t pos,size_t count)
 {
     struct split_raw_private *srp = SPLIT_RAW_PRIVATE(af);
     off_t c1, c3;
@@ -371,7 +386,7 @@ int split_raw_write_internal2(AFFILE *af, unsigned char *buf, uint64 pos,int cou
 	if (af->maxsize) {	// do we need to possibly split into multiple file writes?
 	    /* Figure out which file number we will need to write to... */
 	    if (pos >= (af->maxsize * srp->num_raw_files)) {
-		int fd = open(srp->next_raw_fname, O_RDWR | O_CREAT | O_EXCL, 0666);
+		int fd = open(srp->next_raw_fname, O_RDWR | O_CREAT | O_EXCL | O_BINARY, 0666);
 		if (fd < 0) {
 		  (*af->error_reporter)("split_raw_write: open(%s): ",af->fname);
 		    if (ret) return ret;
@@ -386,11 +401,11 @@ int split_raw_write_internal2(AFFILE *af, unsigned char *buf, uint64 pos,int cou
 			return -1;
 		}
 	    }
-	    i = pos / af->maxsize;
-	    c1 = pos % af->maxsize;
+	    i  = (int)(pos / af->maxsize);
+	    c1 = (off_t)(pos % af->maxsize);
 	} else {
 	    i = 0;
-	    c1 = pos;
+	    c1 = (off_t)pos;
 	}
 	if (c1 != (off_t)srp->pos[i]) {	// do we need to seek this file?
 	    off_t c2 = lseek (srp->fds[i], c1, SEEK_SET); // try to seek
@@ -405,7 +420,7 @@ int split_raw_write_internal2(AFFILE *af, unsigned char *buf, uint64 pos,int cou
 	    srp->pos[i] = c2;
 	}
 	if (af->maxsize && ((af->maxsize - c1) < (unsigned)count))
-	    c3 = af->maxsize - c1;
+	  c3 = (off_t)(af->maxsize - c1);
 	else
 	    c3 = count;
 	if(af->w_callback) {acbi.phase = 3;(*af->w_callback)(&acbi);}
@@ -448,7 +463,7 @@ int split_raw_write_internal2(AFFILE *af, unsigned char *buf, uint64 pos,int cou
     return ret;
 }
 
-int split_raw_write(AFFILE *af, unsigned char *buf, uint64 pos,int count)
+int split_raw_write(AFFILE *af, unsigned char *buf, uint64_t pos,size_t count)
 {
   /* If we are being asked to start writing beyond the end of the file
    * pad out the file (and possibly create one or more new image files.)
@@ -459,10 +474,10 @@ int split_raw_write(AFFILE *af, unsigned char *buf, uint64 pos,int count)
 	  while(pos > af->image_size){	
 
 	      /* repeat until file is as big as where we should be writing */
-	      int64 bytes_left   = pos - af->image_size;
-	      int bytes_to_write = af->maxsize - (af->image_size % af->maxsize); 
-	      if(bytes_to_write > bytes_left) bytes_to_write = bytes_left;
-	      int bytes_written = split_raw_write_internal2(af,0, af->image_size,bytes_to_write); 
+	      int64_t bytes_left   = pos - af->image_size;
+	      int bytes_to_write = (int)(af->maxsize - (af->image_size % af->maxsize)); 
+	      if(bytes_to_write > bytes_left) bytes_to_write = (int)bytes_left;
+	      int bytes_written = split_raw_write_internal2(af,0,af->image_size,bytes_to_write); 
 	      if(bytes_to_write != bytes_written){
 		  return -1;		// some kind of internal error
 	      }
@@ -482,21 +497,21 @@ int split_raw_write(AFFILE *af, unsigned char *buf, uint64 pos,int count)
 static int split_raw_get_seg(AFFILE *af,const char *name,unsigned long *arg,unsigned char *data,
 		       size_t *datalen)
 {
-    int64 page_num = af_segname_page_number(name);
+    int64_t page_num = af_segname_page_number(name);
     if(page_num<0){
 	errno = ENOTSUP;		// sorry! We don't store metadata
 	return -1;
     }
   
-    uint64 pos = page_num * af->image_pagesize; // where we are to start reading
-    uint64 bytes_left = af->image_size - pos;	// how many bytes left in the file
+    uint64_t pos = page_num * af->image_pagesize; // where we are to start reading
+    uint64_t bytes_left = af->image_size - pos;	// how many bytes left in the file
 
     unsigned int bytes_to_read = af->image_pagesize; // copy this many bytes, unless
-    if(bytes_to_read > bytes_left) bytes_to_read = bytes_left; // only this much is left
+    if(bytes_to_read > bytes_left) bytes_to_read = (unsigned int)bytes_left; // only this much is left
     
     if(arg) *arg = 0;			// arg is always 0
     if(datalen){
-	if(*datalen==0 && data==0){ // asked for 0 bytes, so give the actual size
+	if(data==0){ // asked for 0 bytes, so give the actual size
 	    *datalen = bytes_to_read;
 	    return 0;
 	}
@@ -506,8 +521,8 @@ static int split_raw_get_seg(AFFILE *af,const char *name,unsigned long *arg,unsi
 	}
     }
     if(data){
-	unsigned bytes_read = split_raw_read(af,data,pos,bytes_to_read);
-	if(bytes_read==bytes_to_read){
+	int bytes_read = split_raw_read(af,data,pos,bytes_to_read);
+	if(bytes_read>=0){
 	    if(datalen) *datalen = bytes_read;
 	    return 0;
 	}
@@ -527,7 +542,7 @@ static int split_raw_get_next_seg(AFFILE *af,char *segname,size_t segname_len,un
 {
     struct split_raw_private *srp = SPLIT_RAW_PRIVATE(af);
 
-    int64 total_pages = (af->image_size + af->image_pagesize - 1) / af->image_pagesize;
+    int64_t total_pages = (af->image_size + af->image_pagesize - 1) / af->image_pagesize;
     if(srp->cur_page >= total_pages) return -1; // that's all there are
 
     /* Make the segment name */
@@ -560,23 +575,25 @@ static int split_raw_rewind_seg(AFFILE *af)
 }
 
 static int split_raw_update_seg(AFFILE *af, const char *name,
-				unsigned long arg,const void *value,unsigned int vallen,int append)
+				unsigned long arg,const u_char *value,unsigned int vallen)
     
 {
-    int64 page_num = af_segname_page_number(name);
+    int64_t page_num = af_segname_page_number(name);
     if(page_num<0){
 	errno = ENOTSUP;		// sorry! We don't store metadata
 	return -1;
     }
 	
-    uint64 pos = page_num * af->image_pagesize; // where we are to start reading
-    return split_raw_write(af, (unsigned char *)value, pos,vallen);
+    uint64_t pos = page_num * af->image_pagesize; // where we are to start reading
+    int written = split_raw_write(af, (unsigned char *)value, pos,vallen);
+    if(written==vallen) return 0;	// success
+    return -1;
 }
 
 
 struct af_vnode vnode_split_raw = {
     AF_IDENTIFY_SPLIT_RAW,
-    AF_VNODE_TYPE_COMPOUND,
+    AF_VNODE_TYPE_COMPOUND|AF_VNODE_TYPE_RELIABLE|AF_VNODE_MAXSIZE_MULTIPLE|AF_VNODE_NO_SIGNING|AF_VNODE_NO_SEALING,
     "Split Raw",
     split_raw_identify_file,
     split_raw_open,
