@@ -6,26 +6,8 @@
  *         aff -> aff (recompressing/uncompressing)
  */
 
-
-#include "afflib.h"
-#include "afflib_i.h"			// we do enough mucking, we need the internal version
-
-#include <openssl/md5.h>
-#include <openssl/sha.h>
-
-#ifdef UNIX
-#include <unistd.h>
-#include <term.h>
-#endif
-
-#ifdef linux
-#include <sys/time.h>
-#endif
-
-#include <getopt.h>
-
 /*
- * Copyright (c) 2005
+ * Copyright (c) 2005--2008
  *	Simson L. Garfinkel and Basis Technology, Inc. 
  *      All rights reserved.
  *
@@ -66,32 +48,75 @@
 
 
 
+#include "affconfig.h"
+#include "afflib.h"
+#include "afflib_i.h"			// we do enough mucking, we need the internal version
+
+#include <openssl/md5.h>
+#include <openssl/sha.h>
+
+#ifdef WIN32
+#include "unix4win32.h"
+#endif
+
+#ifdef HAVE_CURSES_H
+#include <curses.h>
+#endif
+
+#ifdef HAVE_TERM_H
+#include <term.h>
+#endif
+
+#ifdef HAVE_NCURSES_TERM_H
+#include <ncurses/term.h>
+#endif
+
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
+#ifdef linux
+#include <sys/time.h>
+#endif
+
+#ifdef HAVE_GETOPT_H
+#include <getopt.h>
+#endif
 
 
-char *progname = "afconvert";
+
+const char *progname = "afconvert";
 
 int	image_pagesize = 16*1024*1024;	// default seg size --- 16MB
-int	opt_compress	= 1;
+int	opt_compression_alg	= AF_COMPRESSION_ALG_ZLIB;
 int	opt_compress_level  = AF_COMPRESSION_DEFAULT;
-int64	 bytes_to_convert = 0;
+int64_t	 bytes_to_convert = 0;
 int	opt_batch	= 1;
 int	opt_zap = 0;
 int	opt_quiet	= 0;
 int	opt_nozprobe	= 0;
 int	opt_write_raw		= 0;		// output
-char	*opt_write_raw_ext	= "raw";
-char	*opt_outdir     = 0;
-char	*opt_aff_ext    = "aff";
-int64	opt_maxsize     = 0;
+const char	*opt_write_raw_ext	= "raw";
+const char	*opt_outdir     = 0;
+const char	*opt_aff_ext    = "aff";
+int64_t	opt_maxsize     = 0;
 int	opt_yes		= 0;
 int	opt_debug       = 0;
+char    *command_line = 0;	// what is typed
+
+
+char *append(char *base,const char *str)
+{
+    base = (char *)realloc(base,strlen(base)+strlen(str)+1);
+    strcat(base,str);			// can't fail
+    return base;
+}
+
 
-#define xstr(s) str(s)
-#define str(s) #s
 
 void usage()
 {
-    printf("%s version %s\n",progname,xstr(AFFLIB_VERSION));
+    printf("%s version %s\n",progname,PACKAGE_VERSION);
     printf("\n");
     printf("usage:   %s [options] file1 [... files] \n",progname);
     printf("\n");
@@ -99,6 +124,8 @@ void usage()
     printf("  %s file1.iso --- convert file1.iso to file1.aff\n",progname);
     printf("  %s file1.iso file2.iso file3.iso...  --- batch convert files\n",progname);
     printf("  %s -r -e iso image.aff --- convert image.aff to image.iso\n",progname);
+    printf("  %s -M4g -o/media/dvd.afd  bigfile.aff  --- split an AFF file into 4GB chunks for archiving to DVD\n",
+	   progname);
     //printf("  %s -p image.aff --- recompress image.aff to maximum compression\n",progname);
     printf("\n");
     printf("\nGeneral options:\n");
@@ -107,13 +134,14 @@ void usage()
     printf("\nAFF output options:\n");
     printf("      -a ext   -- use 'ext' for aff files (default is %s)\n",opt_aff_ext);
     printf("                  (use .afd for AFD files)\n");
-    printf("      -M nnn    -- set maximum size of output file. Suffix with g, m or k.\n");
-    printf("      -s nnnn  -- set the image_pagesize (default %d)\n",image_pagesize);
+    printf("      -Mn[kgm] -- set maximum size of output file. Suffix with g, m or k.\n");
+    printf("      -sn      -- set the image_pagesize (default %d)\n",image_pagesize);
     printf("      -x       -- don't compress AFF file.\n");
     printf("      -O dir   -- use 'dir' as the output directory\n");
     printf("      -o file  -- output to 'file' (can only convert one at a time)\n");
     printf("                  File is AFF is file ends .aff; otherwise assumes raw.\n");
-    printf("      -Xn      -- Set compression to n; default is 7");
+    printf("      -Xn      -- Set compression to n; default is 7\n");
+    printf("      -L       -- Use the LZMA compression algorithm (better but slower)\n");
     
     printf("\nRaw output options:\n");
     printf("      -r       -- force raw output. \n");
@@ -124,7 +152,7 @@ void usage()
     printf("      -z       -- zap; delete the output file if it already exists.\n");
     printf("      -Z       -- Do not automatically probe for gzip/bzip2 compression.\n");
     printf("      -y       -- Always answer yes/no questions 'yes.'\n");
-    printf("      -v = Just print the version number and exit.\n");
+    printf("      -V = Just print the version number and exit.\n");
     printf("\n");
     exit(0);
 }
@@ -159,7 +187,7 @@ int probe_bzip2(const char *infile)
  * As a yes/no question. Return 1 if yes, 0 if no.
  */
 
-int yesno(char *statement,char *question,char *affirmative)
+int yesno(const char *statement,const char *question,const char *affirmative)
 {
     if(opt_yes){
 	if(!opt_quiet) printf("%s. %s.\n",statement,affirmative);
@@ -194,13 +222,18 @@ int yesno(char *statement,char *question,char *affirmative)
 #include <vector>
 #include <string>
 
+#ifdef HAVE_CSTRING
+#include <cstring>
+#endif
+
+
 using namespace std;
 
 int convert(const char *infile,char *outfile)
 {
     if(opt_debug) fprintf(stderr,"convert(%s,%s)\n",infile,outfile);
     if(access(infile,F_OK)!=0){
-	err(1,infile);			// file does not exist?
+	err(1,"%s",infile);			// file does not exist?
     }
 
     if(infile && outfile && strcmp(infile,outfile)==0){
@@ -214,12 +247,14 @@ int convert(const char *infile,char *outfile)
      ****************************************************************/
 
     AFFILE *a_in = 0;			// input file, if aff
+#ifdef UNIX
     /* Check to see if it is a gzip file... */
     if(probe_gzip(infile)
        && yesno("infile looks like a gzip file","Uncompress it","Uncompressing")){
 	/* Open with a subprocess. We will need to use zlib when we move to Windows. */
+	if(af_hasmeta(infile)) return -1;	// don't covert with shell metacharacters
 	char buf[256];
-	sprintf(buf,"gzcat %s",infile);
+	snprintf(buf,sizeof(buf),"gzcat %s",infile);
 	a_in = af_popen(buf,"r");
     }
 
@@ -228,16 +263,17 @@ int convert(const char *infile,char *outfile)
        && probe_bzip2(infile)
        && yesno("infile looks like a bzip2 file","Uncompress it","Uncompressing")){
 	/* Open with a subprocess. We will need to use bzip2zlib when we move to Windows. */
+	if(af_hasmeta(infile)) return -1;	// don't covert with shell metacharacters
 	char buf[256];
-	sprintf(buf,"bzcat %s",infile);
+	snprintf(buf,sizeof(buf),"bzcat %s",infile);
 	a_in = af_popen(buf,"r");
     }
-
+#endif
 
     /* If the file isn't open, try to open it... */
     if(!a_in){
 	a_in = af_open(infile,O_RDONLY,0);
-	if(!a_in) err(1,infile);	// give up
+	if(!a_in) af_err(1,"%s",infile);	// give up
 	if(af_identify(a_in)==AF_IDENTIFY_RAW){	
 	    af_set_pagesize(a_in,image_pagesize); // match the page size we want to use
 	}
@@ -251,18 +287,32 @@ int convert(const char *infile,char *outfile)
      *** Open Ouptut
      ****************************************************************/
 
+
     if(opt_zap) unlink(outfile);	// we were told to zap it
 
     AFFILE *a_out = 0;			// output file, if aff or raw...
     if(access(outfile,F_OK)==0){
+	/* If outfile is a device, ask user... */
+	struct stat sb;
+	if(stat(outfile,&sb)){
+	    err(1,"%s exists but can't be stat?",outfile);
+	}
+	if((sb.st_mode & S_IFMT)==S_IFCHR ||
+	   (sb.st_mode & S_IFMT)==S_IFBLK){
+	    char buf[1024];
+	    snprintf(buf,sizeof(buf),"%s is a raw device.\n",outfile);
+	    if(yesno(buf,"Overwrite raw device?","yes")){
+		goto doit;
+	    }
+	}
 	fprintf(stderr,"%s: file exists. Delete it before converting.\n",outfile);
-	return -1;
+	exit(-1);
     }
     /* Check for splitraw names */
     if(af_ext_is(outfile,"afm")){
 	char file000[MAXPATHLEN+1];
-	strcpy(file000,outfile);
-	char *cc = rindex(file000,'.');
+	strlcpy(file000,outfile,sizeof(file000));
+	char *cc = strrchr(file000,'.');
 	if(!cc) err(1,"Cannot file '.' in %s\n",file000);
 	for(int i=0;i<2;i++){
 	    sprintf(cc,".%03d",i);
@@ -274,25 +324,27 @@ int convert(const char *infile,char *outfile)
 	}
     }
 
+ doit:;
+
 
     if(opt_write_raw){
 	/* Easy way to make a raw output is to reopen an existing output file... */
 	FILE *f = fopen(outfile,"w+b");
 	if(!f){
-	    err(1,outfile);
+	    err(1,"%s",outfile);
 	}
 	a_out = af_freopen(f);
-
     }
     else {
 	a_out = af_open(outfile,O_RDWR|O_CREAT|O_BINARY,0777);
+	if(!a_out) af_err(1,"%s",outfile);
 	if(opt_maxsize){
 	    af_set_maxsize(a_out,opt_maxsize);
 	}
 
     }
-    if(a_out == 0) err(1,"af_open: %s",outfile);
-    af_enable_writing(a_out,1);		// we will be writing
+    if(a_out == 0) af_err(1,"af_open: %s",outfile);
+    af_update_seg(a_out,AF_ACQUISITION_COMMAND_LINE,0,(const u_char *)command_line,strlen(command_line));
 
     /****************************************************************
      *** Set up the AFF file (assuming it's an aff file)
@@ -321,13 +373,16 @@ int convert(const char *infile,char *outfile)
     struct af_vnode_info vni;
     af_vstat(a_out,&vni);
     if(vni.supports_compression){
-	af_enable_compression(a_out,
-			      opt_compress ? AF_COMPRESSION_ZLIB : AF_COMPRESSION_NONE,
-			      opt_compress_level);
+	if(opt_compression_alg){
+	    af_enable_compression(a_out,opt_compression_alg,opt_compress_level);
+	}
+	else{
+	    af_enable_compression(a_out,0,0);
+	}
     }
     if(vni.supports_metadata){
-	af_update_seg(a_out,AF_MD5,0,md5_buf,16,1);
-	af_update_seg(a_out,AF_SHA1,0,sha1_buf,20,1);
+	af_update_seg(a_out,AF_MD5,0,md5_buf,16);
+	af_update_seg(a_out,AF_SHA1,0,sha1_buf,20);
     }
 
     /* Get a list of all the metadata segments and the pages
@@ -336,11 +391,11 @@ int convert(const char *infile,char *outfile)
 
     char segname[AF_MAX_NAME_LEN];
     vector <string> metadata_segments;
-    vector <int64> pages;
+    vector <int64_t> pages;
     af_rewind_seg(a_in);			// start at the beginning
-    int64 highest_pagenum = 0;
+    int64_t highest_pagenum = 0;
     while(af_get_next_seg(a_in,segname,sizeof(segname),0,0,0)==0){
-	int64 page_num = af_segname_page_number(segname);
+	int64_t page_num = af_segname_page_number(segname);
 	if(page_num>=0){
 	    pages.push_back(page_num);
 	    if(page_num>highest_pagenum) highest_pagenum = page_num;
@@ -357,7 +412,7 @@ int convert(const char *infile,char *outfile)
 	for(vector<string>::iterator i = metadata_segments.begin();
 	    i != metadata_segments.end();
 	    i++){
-	    strcpy(segname,i->c_str());
+	    strlcpy(segname,i->c_str(),sizeof(segname));
 	    size_t data_len = 0;
 	    unsigned long arg;
 
@@ -374,7 +429,7 @@ int convert(const char *infile,char *outfile)
 		continue;
 	    }
 	    /* Now put the data */
-	    if(af_update_seg(a_out,segname,arg,data,data_len,1)){
+	    if(af_update_seg(a_out,segname,arg,data,data_len)){
 		err(1,"af_update_seg");
 	    }
 	    free(data);
@@ -387,31 +442,29 @@ int convert(const char *infile,char *outfile)
      */
     sort(pages.begin(),pages.end());
     
-    int64  prev_pagenum = -1;
+    int64_t  prev_pagenum = -1;
     bool   hash_valid = true;
-    uint64 last_byte_in_image = 0;
-    uint64 total_bytes_converted = 0;
+    uint64_t last_byte_in_image = 0;
+    uint64_t total_bytes_converted = 0;
 
     bool copy_by_pages = af_has_pages(a_in);
 
     if(copy_by_pages){
 	/* Copy over data one page at a time */
-	for(vector<int64>::iterator i = pages.begin();
-	    i != pages.end();
-	    i++){
+	for(vector<int64_t>::iterator i = pages.begin(); i != pages.end(); i++){
 	    
-	    int64 pagenum = *i;
+	    int64_t pagenum = *i;
 	    
 	    if(!opt_quiet) printf("Converting page %"I64d" of %"I64d"\r",pagenum,highest_pagenum);fflush(stdout);
 	    
 	    unsigned char *data = (unsigned char *)malloc(image_pagesize);
 	    size_t data_len = image_pagesize;
 	    if(af_get_page(a_in,pagenum,data,&data_len)){
-		err(1,"af_get_page(file=%s,page=%"I64d,
+		err(1,"af_get_page(file=%s,page=%"I64d")",
 		    af_filename(a_in),pagenum);
 	    }
 	    if(af_update_page(a_out,pagenum,data,data_len)){
-		err(1,"af_update_page(file=%s,page=%"I64d,
+		err(1,"af_update_page(file=%s,page=%"I64d")",
 		    af_filename(a_out),pagenum);
 	    }
 	    
@@ -423,7 +476,7 @@ int convert(const char *infile,char *outfile)
 		prev_pagenum = pagenum;
 	    }
 	    free(data); data = 0;
-	    last_byte_in_image = (int64)image_pagesize * pagenum + (int64)data_len;
+	    last_byte_in_image = (int64_t)image_pagesize * pagenum + (int64_t)data_len;
 	    total_bytes_converted += data_len;
 	}
     } else {
@@ -460,36 +513,37 @@ int convert(const char *infile,char *outfile)
     if(hash_valid){
 	MD5_Final(md5_buf,&md5);
 	char buf[256];
-	if(af_update_seg(a_out,AF_MD5,0,md5_buf,16,1) && errno!=ENOTSUP){
+	if(af_update_seg(a_out,AF_MD5,0,md5_buf,16) && errno!=ENOTSUP){
 	    err(1,"Could not update AF_MD5");
 	}
 	if(!opt_quiet) printf("md5: %s\n",af_hexbuf(buf,sizeof(buf),md5_buf,16,1));
 	
 	SHA1_Final(sha1_buf,&sha);
-	if(af_update_seg(a_out,AF_SHA1,0,sha1_buf,20,1) && errno!=ENOTSUP){
+	if(af_update_seg(a_out,AF_SHA1,0,sha1_buf,20) && errno!=ENOTSUP){
 	    err(1,"Could not update AF_SHA1");
 	}
 	if(!opt_quiet) printf("sha1: %s\n",af_hexbuf(buf,sizeof(buf),sha1_buf,20,1));
     }
 
     /* Go back and update the image size */
-    if(af_update_segq(a_out,AF_IMAGESIZE,last_byte_in_image,1)
+    if(af_update_segq(a_out,AF_IMAGESIZE,last_byte_in_image)
        && errno!=ENOTSUP){
 	err(1,"Could not upate AF_IMAGESIZE");
     }
 
     /* Finish the hash calculations and write to the db */
     if(!opt_quiet){
-	printf("bytes converted: %qd \n",total_bytes_converted);
+	printf("bytes converted: %"I64d" \n",total_bytes_converted);
 
 	/* If the vnode implementation tracked segments written, report it. */
-	if(a_out->usegs_written || a_out->csegs_written){
-	    printf("Total segments: %u  (%u compressed)\n",
-		   a_out->usegs_written+a_out->csegs_written,
-		   a_out->csegs_written);
+	if(a_out->pages_written || a_out->pages_compressed){
+	    printf("Total pages: %"I64u"  (%"I64u" compressed)\n",
+		   a_out->pages_written,a_out->pages_compressed);
 	}
     }
 
+    af_make_gid(a_out);			// make an AF_IMAGE_GID if it doesn't exist
+
     if(af_close(a_in)) err(1,"af_close(a_in)");
     if(af_close(a_out)) err(1,"af_close(a_out)");
 
@@ -499,18 +553,18 @@ int convert(const char *infile,char *outfile)
     }
 
     /* Set the utime on the resulting file if we can stat it */
-    return(0);
+    exit(0);
 }
 
 
-int64 atoi64(const char *buf)
+int64_t atoi64(const char *buf)
 {
-    int64 r=0;
+    int64_t r=0;
     sscanf(buf,"%"I64d,&r);
     return r;
 }
 
-int64 atoi64m(const char *optarg)
+int64_t atoi64m(const char *optarg)
 {
     int multiplier;
     switch(optarg[strlen(optarg)-1]){
@@ -538,9 +592,21 @@ int main(int argc,char **argv)
     char *outfile = 0;
     int ch;
 
+#ifdef HAVE_LIBNCURSES
     setupterm((char *)0,1,(int *)0);
+#endif
 
-    while ((ch = getopt(argc, argv, "a:e:o:zqrs:xX:Zh?M:O::ydv")) != -1) {
+    /* Figure out the command line that was used to run this program.
+     * This information will be recorded in the AFF file.
+     */
+    command_line = (char *)malloc(1);
+    command_line[0] = 0;
+    for(int i=0;i<argc;i++){
+	if(i>0) command_line = append(command_line," ");
+	command_line = append(command_line,argv[i]);
+    }
+
+    while ((ch = getopt(argc, argv, "a:e:Lo:zqrs:xX:Zh?M:O::ydV")) != -1) {
 	switch (ch) {
 	case 'a':
 	    opt_aff_ext = optarg;
@@ -558,6 +624,9 @@ int main(int argc,char **argv)
 	case 'q':
 	    opt_quiet++;
 	    break;
+	case 'L':
+	    opt_compression_alg = AF_COMPRESSION_ALG_LZMA;
+	    break;
 	case 'r':
 	    opt_write_raw++;
 	    break;
@@ -565,10 +634,9 @@ int main(int argc,char **argv)
 	    image_pagesize = atoi64m(optarg);
 	    break;
 	case 'x':
-	    opt_compress=0;
+	    opt_compression_alg=AF_COMPRESSION_ALG_NONE;
 	    break;
 	case 'X':
-	    opt_compress=1;
 	    opt_compress_level = atoi(optarg);
 	    break;
 	case 'Z':
@@ -592,8 +660,8 @@ int main(int argc,char **argv)
 	default:
 	    usage();
 	    exit(0);
-	case 'v':
-	    printf("%s version %s\n",progname,xstr(AFFLIB_VERSION));
+	case 'V':
+	    printf("%s version %s\n",progname,PACKAGE_VERSION);
 	    exit(0);
 	    
 	}
@@ -612,53 +680,53 @@ int main(int argc,char **argv)
 
     while(*argv){
 	/* Come up with output file name */
-	char outfile[PATH_MAX];
+	char outfile[MAXPATHLEN+1];
 	memset(outfile,0,sizeof(outfile));
 
-	char *ext = opt_write_raw ? opt_write_raw_ext : opt_aff_ext;
+	const char *ext = opt_write_raw ? opt_write_raw_ext : opt_aff_ext;
 	char *infile = *argv;
 	argv++;
 	argc--;
 
 	/* Check for "-o filename" to specify the output arg... */
 	if(argc>=2 && strcmp(*argv,"-o")==0){
-	    strcpy(outfile,argv[1]);
+	    strlcpy(outfile,argv[1],sizeof(outfile));
 	    argv += 2;
 	    argc -= 2;
 	}
 	else {
 
 	    /* Copy over the filename and change the extension */
-	    strcpy(outfile,infile);
-	    char *cc = rindex(outfile,'.'); // to strip off extension
+	    strlcpy(outfile,infile,sizeof(outfile));
+	    char *cc = strrchr(outfile,'.'); // to strip off extension
 	    if(cc){
 		/* Found an extension; copy over mine. */
-		strcpy(cc+1,ext);
+		strlcpy(cc+1,ext,sizeof(outfile)-(cc-outfile));
 	    }
 	    else {
 		/* No extension; make one */
-		strcat(outfile,".");
-		strcat(outfile,ext);
+		strlcat(outfile,".",sizeof(outfile));
+		strlcat(outfile,ext,sizeof(outfile));
 	    }
 
 	    /* The user might want us to put things
 	     * in a different directory. Pull off the filename...
 	     */
 	    if(opt_outdir){
-		cc = rindex(outfile,'/');
+		cc = strrchr(outfile,'/');
 		char filename[PATH_MAX];
 		if(cc){
-		    strcpy(filename,cc+1);	// just the filename
+		    strlcpy(filename,cc+1,sizeof(filename));	// just the filename
 		}
 		else{
-		    strcpy(filename,outfile);	// the outfile is the filename
+		    strlcpy(filename,outfile,sizeof(filename));	// the outfile is the filename
 		}
-		strcpy(outfile,opt_outdir);
-		strcat(outfile,"/");
-		strcat(outfile,filename);
+		strlcpy(outfile,opt_outdir,sizeof(outfile));
+		strlcat(outfile,"/",sizeof(outfile));
+		strlcat(outfile,filename,sizeof(outfile));
 	    }
 	}
 	convert(infile,outfile);
     }
-    return 0;
+    exit(0);
 }
