@@ -5,7 +5,7 @@
  */
 
 /*
- * Copyright (c) 2005
+ * Copyright (c) 2005--2008
  *	Simson L. Garfinkel and Basis Technology, Inc. 
  *      All rights reserved.
  *
@@ -45,10 +45,16 @@
  */
 
 
+#include "affconfig.h"
 #include "afflib.h"
+#include "utils.h"
 #include "afflib_i.h"
 #include "quads.h"
 
+#ifdef USE_S3
+#include "s3_glue.h"
+#endif
+
 #include <ctype.h>
 #include <zlib.h>
 #include <openssl/md5.h>
@@ -59,20 +65,30 @@
 #include <cstdlib>
 #include <vector>
 #include <string>
+
+#ifdef HAVE_CSTRING
+#include <cstring>
+#endif
+
 using namespace std;
 
-#ifdef UNIX
-#include <unistd.h>
+#ifdef HAVE_CURSES_H
+#include <curses.h>
+#endif
+
+#ifdef HAVE_TERM_H
 #include <term.h>
-#define HAS_CURSES
+#endif
+
+#ifdef HAVE_NCURSES_TERM_H
+#include <ncurses/term.h>
 #endif
 
 #ifdef WIN32
 #include "unix4win32.h"
-#include <malloc.h>
 #endif
 
-char *progname = "afinfo";
+const char *progname = "afinfo";
 
 #define VALIDATE_MD5  0x01
 #define VALIDATE_SHA1 0x02
@@ -81,6 +97,7 @@ int opt_validate = 0;
 int opt_info = 1;
 int opt_all  = 0;
 int opt_wide = 0;
+int opt_l = 0;
 unsigned int cols = 80;				// default
 int opt_x = 0;
 int opt_b = 0;
@@ -89,16 +106,56 @@ int opt_verbose = 0;
 int opt_y = 0;
 int opt_hexbuf = AF_HEXBUF_SPACE4 | AF_HEXBUF_UPPERCASE;
 int opt_page_validate = 0;
-
-#define xstr(s) str(s)
-#define str(s) #s
+int opt_no_preview = 0;
+int opt_preview_md5 = 0;
+int opt_debug = 0;
+int opt_figure_media = 0;
+const char *opt_passphrase = 0;
 
 vector<string> opt_seglist;		// just info these segments
+bool something_was_decrypted = false;
+
+
+/**
+ * select bold on or off
+ */
+void bold(int on)
+{
+#if defined(HAVE_LIBNCURSES)
+#ifdef HAVE_ISATTY
+    if(!isatty(fileno(stdout))) return;
+#endif
+    if(on) tputs(enter_bold_mode,1,putchar);
+    else tputs(exit_attribute_mode,0,putchar);
+#endif
+}
+
+/**
+ * select a color.
+ * @param num - 0 is black; 1 red; 2 green; 3 yellow; 4 blue; 5 magenta; 6 cyan; 7 white;
+ */
+  
+#define RED 1
+#define WHITE 7
+
+void color(int num)
+{
+#ifdef HAVE_ISATTY
+    if(!isatty(fileno(stdout))) return;
+#endif
+#ifdef HAVE_CURSES_H
+    char *setf = tigetstr("setf");
+    if(!setf) setf = tigetstr("setaf");
+    if(setf){
+	putp(tparm(setf,num));
+    }
+#endif
+}
 
 
 void usage()
 {
-    printf("%s version %s\n",progname,xstr(AFFLIB_VERSION));
+    printf("%s version %s\n",progname,PACKAGE_VERSION);
     printf("usage: %s [options] infile\n",progname);
     printf("   -a = print ALL segments (normally data segments are suppressed)\n");
     printf("   -b = print how many bad blocks in each segment (implies -a)\n");
@@ -106,12 +163,28 @@ void usage()
     printf("   -w = wide output; print more than 1 line if necessary.\n");
     printf("   -s segment =   Just print information about 'segment'.\n");
     printf("                    (may be repeated)\n");
-    printf("   -x = print binary values in hex (default is ASCII)\n");
     printf("   -m = validate MD5 hash of entire image\n");
     printf("   -S = validate SHA1 hash of entire image\n");
-    printf("   -p = validate the hash of each page (if present)\n");
+    printf("   -v = validate the hash of each page (if present)\n");
     printf("   -y = don't print segments of lengths 16 and 20 as hex)\n");
-    printf("   -v = Just print the version number and exit.\n");
+    printf("   -p<passphrase> = Specify <passphrase> to decrypt file\n");
+    printf("   -l = Just print the segment names and exit\n");
+    printf("   -V = Just print the version number and exit.\n");
+
+    printf("\nPreview Options:\n");
+    printf("   -X = no data preview; just print the segment names\n");
+    printf("   -x = print binary values in hex (default is ASCII)\n");
+    printf("\nMisc:\n");
+    printf("   -d = debug\n");
+    printf("   -A = if infile is a device, print the number of sectors\n");
+    printf("        and sector size to stdout in XML. Otherwise error\n");
+
+    if(opt_debug){
+	for(int i=0;i<9;i++){
+	    color(i);printf("Color %d\n",i);color(7);
+	}
+    }
+
     exit(0);
 }
 
@@ -130,6 +203,7 @@ void sig_info(int arg)
 void validate(const char *infile)
 {
     af = af_open(infile,O_RDONLY,0);
+    if(!af) af_err(1,infile);
     switch(af_identify(af)){
     case AF_IDENTIFY_AFF:
     case AF_IDENTIFY_AFM:
@@ -161,6 +235,12 @@ void validate(const char *infile)
 	int page_num = af_segname_page_number(segname);
 	if(page_num>=0) pages.push_back(page_num);
     }
+
+    if(pages.size()==0){
+	printf("No pages to validate.\n");
+	af_close(af);
+    }
+
     sort(pages.begin(),pages.end());
     vector<int>::iterator i = pages.begin();
     int last = *i;
@@ -181,7 +261,7 @@ void validate(const char *infile)
     SHA_CTX sha;
     SHA1_Init(&sha);
     
-    uint64 total_bytes = 0;
+    uint64_t total_bytes = 0;
     while(!af_eof(af)){
 	unsigned char buf[65536];		// a decent size
 	size_t bytes = af_read(af,buf,sizeof(buf));
@@ -258,9 +338,11 @@ bool display_as_hex(const char *segname,int data_len)
 {
     if(strcmp(segname,AF_MD5)==0) return true;
     if(strcmp(segname,AF_SHA1)==0) return true;
+    if(strcmp(segname,AF_IMAGE_GID)==0) return true;
     if(data_len==16 && strstr(segname,"md5")) return true;
     if(data_len==20 && strstr(segname,"sha1")) return true;
     if(opt_x) return true;
+    if(opt_preview_md5) return true;
     return false;
 }
 
@@ -277,7 +359,7 @@ void badscan(AFFILE *af,int page_number,size_t data_len)
     int funny_sectors = 0;
     for(unsigned int offset=0;offset<page_size;offset+=af->image_sectorsize){
 	sectors++;
-	if(af_is_badblock(af,buf+offset)){
+	if(af_is_badsector(af,buf+offset)){
 	    bad_sectors ++;
 	    continue;
 	}
@@ -304,15 +386,22 @@ void badscan(AFFILE *af,int page_number,size_t data_len)
 void print_info(AFFILE *af,const char *segname)
 {
     unsigned long arg;
+    unsigned char *data = 0;
+    int dots = 0;
+    uint display_len = 0;
+    char *cc = 0;
 
     /* Check to see if this is a null page. */
     if(segname[0]==0 && opt_all==0){
 	return;
     }
+    if(segname[0]=='[' && opt_all){
+	puts(segname);
+	return;
+    }
 
     /* Check to see if this is a data page */
-    int64 page_num = af_segname_page_number(segname);
-    if(page_num>=0 && opt_all==0) return;		// don't print data pages?
+    int64_t page_num = af_segname_page_number(segname);
 	
     size_t data_len = 0;
     /* First find out how big the segment is, then get the data */
@@ -320,18 +409,37 @@ void print_info(AFFILE *af,const char *segname)
 	printf("%-25s  SEGMENT NOT FOUND\n",segname);
 	return;
     }
-    unsigned char *data = (unsigned char *)malloc(data_len);
-    if(af_get_seg(af,segname,0,data,&data_len)){
-	warn("af_get_seg_2");
-	free(data);
-	return;
+
+    /* is this an encrypted segment that I have decrypted?
+     * Turn off automatic decryption and see if I can get it again...
+     * If we can get it again, then it wasn't decrypted.
+     */
+    af_set_option(af,AF_OPTION_AUTO_DECRYPT,0);
+    bool was_decrypted = ( af_get_seg(af,segname,0,0,0)!=0) ;
+    af_set_option(af,AF_OPTION_AUTO_DECRYPT,1);    
+
+    if(was_decrypted){
+	bold(1);
+	something_was_decrypted = true;	// print key at bottom
     }
 
+    /* Start the output line */
     char output_line[OUTLINE_LEN];		
     memset(output_line,0,sizeof(output_line));
 
     /* Now append the arg and the data len */
-    sprintf(output_line,"%-20s %6ld   %6ld   ",segname,arg,(long)data_len);
+    sprintf(output_line,"%-24s %8ld   %8ld   ",segname,arg,(long)data_len);
+
+    if(opt_no_preview){
+	printf("%s\n",output_line);
+	goto done;
+    }
+
+    data = (unsigned char *)malloc(data_len);
+    if(af_get_seg(af,segname,0,data,&data_len)){
+	warn("af_get_seg_2 failed: segname=%s data_len=%d",segname,data_len);
+	goto done;
+    }
 
     /* Special handling of values that should be displayed as time */
     if(display_as_time(segname)){
@@ -339,8 +447,7 @@ void print_info(AFFILE *af,const char *segname)
 	int minutes = (arg / 60) % 60;
 	int seconds = arg % 60;
 	printf("%s= %02d:%02d:%02d (hh:mm:ss)\n",output_line,hours,minutes,seconds);
-	free(data);
-	return;
+	goto done;
     }
 
     /* Special handling of quadwords that should be printed as such? */
@@ -360,31 +467,35 @@ void print_info(AFFILE *af,const char *segname)
 	default:
 	    printf("%s= CANNOT DECODE %d byte segment\n",output_line,(int)data_len);
 	}
-	free(data);
-	return;
+	goto done;
     }
 
 
     /* See if I need to truncate */
-    int dots = 0;
-    uint display_len = data_len;
+    display_len = data_len;
     if(opt_wide==0 && data_len>32){ // don't bother showing more than first 32 bytes
 	dots = 1;
 	display_len = 32;
     }
 	    
-    char *cc = output_line + strlen(output_line);
+    cc = output_line + strlen(output_line);
 
+    if(opt_preview_md5){
+	u_char md5[32];
+	MD5(data,data_len,md5);
+	memcpy(data,md5,32);
+	data_len = 32;
+    }
     if(display_as_hex(segname,display_len)){
 	char buf[80];
-	sprintf(cc,"%s%s",af_hexbuf(buf,sizeof(buf),data,display_len,opt_hexbuf),
+	snprintf(cc,sizeof(output_line)-strlen(output_line),
+		 "%s%s",af_hexbuf(buf,sizeof(buf),data,display_len,opt_hexbuf),
 		dots ? "..." : "");
 	/* Special code for SHA1 */
 	if(!opt_wide && strcmp(segname,AF_SHA1)==0){
-	    fwrite(output_line,1,78,stdout);
-	    printf("\n%49s\n",output_line+78);
-	    free(data);
-	    return;
+	    fwrite(output_line,1,82,stdout);
+	    printf("\n%62s\n",output_line+82);
+	    goto done;
 	}
     }
     else {
@@ -438,7 +549,8 @@ void print_info(AFFILE *af,const char *segname)
 	MD5(page_data,page_data_len,hash_calc);
 	if(memcmp(hash_buf,hash_calc,sizeof(hash_buf))!=0){
 	    char hb[32];
-	    printf("** HASH INVALID **\n%30s Calculated %s\n","",af_hexbuf(hb,sizeof(hb),hash_calc,16,opt_hexbuf));
+	    printf("** HASH INVALID **\n%30s Calculated %s\n","",
+		   af_hexbuf(hb,sizeof(hb),hash_calc,16,opt_hexbuf));
 	    printf("%30s Wanted %s ","",af_hexbuf(hb,sizeof(hb),hash_buf,16,opt_hexbuf));
 	    printf("data_len=%d\n",(int)data_len);
 	} else{
@@ -448,36 +560,77 @@ void print_info(AFFILE *af,const char *segname)
     }
  skip1:;
     putchar('\n');
-
-    free(data);
+ done:
+    if(data) free(data);
+    bold(0);			// make sure bold is off
+    //color(WHITE);		// make sure we are back to normal color
 }
 
 
+/* Print the information on a specific file. */
 int info_file(const char *infile)
 {
     unsigned long total_segs = 0;
     unsigned long total_pages = 0;
-    unsigned long null_segs = 0;
-
+    unsigned long total_hashes = 0;
+    unsigned long total_signatures =0;
+    unsigned long total_nulls = 0;
+    struct af_vnode_info vni;
 
     AFFILE *af = af_open(infile,O_RDONLY,0);
-    
-    if(!af){
-	err(1,"Cannot open %s",infile);
+    if(!af) af_err(1,"Cannot open %s",infile);
+    if(af_vstat(af,&vni)) err(1,"%s: af_vstat failed",infile);
+
+    if(opt_l){
+	/* Just list the segments and exit */
+        aff::seglist sl;
+	sl.get_seglist(af);
+	for(aff::seglist::const_iterator i = sl.begin(); i!=sl.end(); i++){
+	    printf("%s\n",(*i).name.c_str());
+	}
+	af_close(af);
+	return 0;
     }
 
-    printf("\n%s\n",af_filename(af));
-    if(opt_all==0){
-	printf("[skipping data segments]\n");
+    if(vni.segment_count_encrypted>0 || vni.segment_count_signed>0){
+	printf("%s: has %s%s%ssegments\n",infile,
+	       (vni.segment_count_encrypted ? "encrypted " : ""),
+	       ((vni.segment_count_encrypted && vni.segment_count_signed) ? "and ": ""),
+	       (vni.segment_count_signed ? "signed " : ""));
     }
-    printf("                                 data       \n");
-    printf("Segment                 arg    length   data\n");
-    printf("=======             =======    ======   =====\n");
+
+
+    if(opt_passphrase){
+	if(af_use_aes_passphrase(af,opt_passphrase)){
+	    errx(1,"%s: cannot use passphrase",opt_passphrase);
+	}
+    }
+    
+    printf("\n%s\n",af_filename(af));
+    const char *v1 = "data";
+    const char *v2 = "====";
+
+    if(opt_all==0) printf("[skipping data segments]\n");
+    if(opt_all==0 && vni.segment_count_encrypted) printf("[skipping encrypted segments]\n");
+    if(opt_no_preview){
+	v1 = "";
+	v2 = "";
+    }
+    if(opt_preview_md5){
+	v1 = "md5";
+    }
+
+    printf("                                        data       \n");
+    printf("Segment                       arg      length    %s\n",v1);
+    printf("=======                 =========    ========    %s\n",v2);
 
     /* If a list of segments was specified by the user, just use that list */
     if(opt_seglist.size()>0){
-	for(vector<string>::iterator i = opt_seglist.begin(); i != opt_seglist.end(); i++){
-	    print_info(af,i->c_str());
+	for(vector<string>::iterator i = opt_seglist.begin();
+	    i != opt_seglist.end();
+	    i++){
+	    const char *segname = i->c_str();
+	    print_info(af,segname);
 	}
 	af_close(af);
 	return 0;
@@ -487,51 +640,166 @@ int info_file(const char *infile)
     vector <string> segments;
     char segname[AF_MAX_NAME_LEN];
     af_rewind_seg(af);			// start at the beginning
-    while(af_get_next_seg(af,segname,sizeof(segname),0,0,0)==0){
-	int page_num = af_segname_page_number(segname);
-	if(page_num>=0) total_pages++;
-	segments.push_back(segname);
-	if(segname[0]==0) null_segs++;
+    int64_t total_datalen = 0;
+    size_t total_segname_len = 0;
+    size_t datalen = 0;
+    int aes_segs=0;
+    while(af_get_next_seg(af,segname,sizeof(segname),0,0,&datalen)==0){
 	total_segs ++;
+	total_datalen += datalen;
+	total_segname_len += strlen(segname);
+	if(segname[0]==0) total_nulls++;
+
+	/* Check to see if this is a regular page or a hash page */
+	char hash[64];
+	int64_t page_num = af_segname_page_number(segname);
+	int64_t hash_num = af_segname_hash_page_number(segname,hash,sizeof(hash));
+	if(page_num>=0) total_pages++;
+	if(hash_num>=0) total_hashes++;
+	if(strstr(segname,AF_SIG256_SUFFIX)) total_signatures++;
+	if(strstr(segname,AF_AES256_SUFFIX)) aes_segs++;
+	if(opt_all==0 && (page_num>=0||hash_num>=0)) continue;	// skip
+	if(opt_all==0 && af_is_encrypted_segment(segname)) continue; // skip
+
+	if(segname[0]==0 && datalen>0 && opt_all){
+	    sprintf(segname,"[null %d bytes]",datalen);
+	}
+	segments.push_back(segname);
     }
 
     /* Now process the segments */
-    for(vector<string>::iterator i = segments.begin(); i != segments.end(); i++){
+    for(vector<string>::const_iterator i = segments.begin();
+	i != segments.end(); i++){
 	print_info(af,i->c_str());
     }
+
+    /* Print the key */
+    if(something_was_decrypted){
+	bold(1);
+	printf("Bold indicates segments that were decrypted.\n");
+	bold(0);
+    }
+
+
     printf("\n");
-    printf("Page  segments:    %8lu\n",total_pages);
-    printf("Empty segments:    %8lu\n",null_segs);
-    printf("Total segments:    %8lu\n", total_segs);
-    //printf("Compression ratio: %5.2g\n", total_bytes_compressed / total_bytes_uncompressed
+    printf("Total segments:        %8lu   (%d real)\n", total_segs,total_segs-total_nulls);
+    if(aes_segs){
+	printf("  Encrypted segments:  %8lu\n",aes_segs);
+    }
+    printf("  Page  segments:      %8lu\n",total_pages);
+    printf("  Hash  segments:      %8lu\n",total_hashes);
+    printf("  Signature segments:  %8lu\n",total_signatures);
+    printf("  Null segments:       %8lu\n",total_nulls);
+    if(opt_all){
+	printf("  Empty segments:      %8lu\n",total_nulls);
+	printf("\n");
+	printf("Total data bytes in segments: %"I64d"\n",total_datalen);
+
+	printf("Total space in file dedicated to segment names: %d\n",
+	       total_segname_len);
+	printf("Total overhead for %d segments: %d bytes (%d*(%d+%d))\n",
+	   total_segs,
+	   total_segs*(sizeof(struct af_segment_head)
+			   +sizeof(struct af_segment_tail)),
+	   total_segs,sizeof(struct af_segment_head),
+	   sizeof(struct af_segment_tail));
+	printf("Overhead for AFF file header: %d bytes\n",sizeof(struct af_head));
+    }
+
+    int64_t device_sectors = 0;
+    af_get_segq(af,AF_DEVICE_SECTORS,&device_sectors);
+    if(device_sectors==0){
+	/* See if we can fake it */
+	unsigned long cylinders=0;
+	unsigned long heads=0;
+	unsigned long sectors_per_track=0;
+	af_get_seg(af,AF_CYLINDERS,&cylinders,0,0);
+	af_get_seg(af,AF_HEADS,&heads,0,0);
+	af_get_seg(af,AF_SECTORS_PER_TRACK,&sectors_per_track,0,0);
+	device_sectors = cylinders * heads * sectors_per_track;
+    }
+    //printf("device_sectors=%"I64d"\n",device_sectors);
+
+
+    int some_missing_pages = 1;
+    if(af->image_pagesize && af->image_sectorsize && device_sectors){
+	int64_t device_bytes = (int64_t)device_sectors * af->image_sectorsize;
+	int64_t device_pages = (device_bytes+af->image_pagesize-1) / af->image_pagesize;
+	int64_t missing_pages = device_pages - total_pages;
+	//printf("device_bytes=%"I64d"\n",device_bytes);
+	//printf("device_pages=%"I64d"\n",device_pages);
+	if(missing_pages!=0){
+	    printf("Missing page segments: %8"I64u"\n",missing_pages);
+	}
+	else {
+	    some_missing_pages=0;
+	}
+    }
+    if (some_missing_pages){
+	if(((total_pages-1) * af->image_pagesize <= af->image_size) &&
+	   ((total_pages)   * af->image_pagesize >= af->image_size)){
+	    some_missing_pages = 0;
+	}
+    }
+
+    if(some_missing_pages && opt_debug){
+	printf("Cannot calculate missing pages\n");
+	printf("  device_sectors=%"I64d" image_pagesize=%d sectorsize=%d\n",
+	device_sectors,af->image_pagesize,af->image_sectorsize);
+    }
     af_close(af);
     return 0;
 
 }
 	 
 
+void figure_media(const char *fn)
+{
+    int fd = open(fn,O_RDONLY,0);
+    if(fd<0) err(1,"open(%s)",fn);
+    struct af_figure_media_buf afb;
+    if(af_figure_media(fd,&afb)){
+	err(1,"af_figure_media(%s)",fn);
+    }
+    printf("<?xml version='1.0' encoding='UTF-8'?>\n");
+    printf("<!DOCTYPE Server >\n");
+    printf("<device name='%s'>\n",fn);
+    printf("   <sector_size>%d</sector_size>\n",afb.sector_size);
+    printf("   <total_sectors>%"PRId64"</total_sectors>\n",afb.total_sectors);
+    printf("   <max_read_blocks>%d</max_read_blocks>\n",afb.max_read_blocks);
+    printf("</device>\n");
+    close(fd);
+}
+
 int main(int argc,char **argv)
 {
     int ch;
     const char *infile;
 
     /* Figure out how many cols the screen has... */
-#ifdef HAS_CURSES
+#ifdef HAVE_LIBNCURSES
     setupterm((char *)0,1,(int *)0);
+    start_color();
     cols = tgetnum("co");
 #endif
 
-    while ((ch = getopt(argc, argv, "abh?s:SmiIwj:pxv")) != -1) {
+    while ((ch = getopt(argc, argv, "abh?s:SmiIwj:p:xvVX5dAl")) != -1) {
 	switch (ch) {
 	case 'a': opt_all++; break;
 	case 'b': opt_all ++; opt_b   ++; break;
 	case 'i': opt_info=0; opt_identify = 1; break;
 	case 'w': opt_wide++; break;
+	case 'X': opt_no_preview++;break;
 	case 'x': opt_x++; break;
 	case 'y': opt_y++; break;
+	case 'l': opt_l++;break;
 	case 'm': opt_validate |= VALIDATE_MD5; break;
 	case 'S': opt_validate |= VALIDATE_SHA1; break;
-	case 'p': opt_page_validate = 1;break;
+	case 'v': opt_page_validate = 1;break;
+	case 'p': opt_passphrase = optarg; break;
+	case '5': opt_preview_md5 = 1;break;
+	case 'd': opt_debug = 1;break;
+	case 'A': opt_figure_media = 1 ; break;
 
 	case 'h':
 	case '?':
@@ -541,8 +809,8 @@ int main(int argc,char **argv)
 	case 's':
 	    opt_seglist.push_back(optarg); // add to the list of segments to info
 	    break;
-	case 'v':
-	    printf("%s version %s\n",progname,xstr(AFFLIB_VERSION));
+	case 'V':
+	    printf("%s version %s\n",progname,PACKAGE_VERSION);
 	    exit(0);
 	}
     }
@@ -559,15 +827,18 @@ int main(int argc,char **argv)
 	infile = *argv++;		// get the file
 	argc--;				// decrement argument counter
 	 
-	const char *name = af_identify_name(infile);
-	if(!name) err(1,infile);
+	const char *name = af_identify_file_name(infile,1);
+	if(!name) err(1,"%s",infile);
 
+	if(opt_figure_media){ figure_media(infile);continue;}
 	if(opt_identify) printf("%s is a %s file\n",infile,name);
-	 
 	if(opt_info)	 info_file(infile);
 	if(opt_validate) validate(infile);
     }
-    return 0;
+#ifdef USE_S3
+    s3_audit(0);
+#endif
+    return(0);
 }
 
 
